import { User } from 'entities';
import { FastifyRequest } from 'fastify';
import axios from 'axios';
import * as argon2 from 'argon2';
import { ForbiddenException, UnauthorizedException } from '@nestjs/common';
import TfaAuthService from './tfa';
import { UserRole } from 'dto';

const GITLAB_ID =
	'98bcbad78cb1f880d1d1de62291d70a791251a7bea077bfe7df111ef3c115760';
const GITLAB_SECRET =
	'7ee01d2b204aff3a05f9d028f004d169b6d381ec873e195f314b3935fa150959';
const GITLAB_URL = 'https://gitlab.mattv.de';
const GITLAB_API_URL = 'https://ssh.gitlab.mattv.de';

interface GitlabTokenResponse {
	access_token: string;
	token_type: string;
	expires_in: number;
	refresh_token: string;
	created_at: number;
}

interface GitlabUserResponse {
	username: string;
	is_admin?: boolean;
}

export default class GitlabAuthService extends TfaAuthService {
	async login(req: Request, user: User) {
		if (user.isGitlabUser && !(await this.verifyGitlabUser(req, user))) {
			await this.revokeAll(user);
			throw new UnauthorizedException('Invalid gitlab token');
		}
		return super.login(req, user);
	}

	async singupGitlab(
		info: GitlabUserResponse,
		data: GitlabTokenResponse
	): Promise<User> {
		const user = new User();
		user.name = info.username;
		user.password = '';
		user.isGitlabUser = true;
		user.role = info.is_admin ? UserRole.ADMIN : UserRole.DISABLED;
		return this.singupInternal(await this.setGitlabTokens(user, data));
	}

	async setGitlabTokens(
		user: User,
		data: GitlabTokenResponse
	): Promise<User> {
		user.gitlabAT = data.access_token;
		user.gitlabRT = data.refresh_token;
		return this.userRepo.save(user);
	}

	getGitlabRedirectUrl(req: Request): string {
		const _req = req as unknown as FastifyRequest;
		return `${_req.protocol}://${_req.hostname}/api/auth/gitlab_callback`;
	}

	getGitlabAuthUrl(req: Request): string {
		const params = new URLSearchParams();
		params.append('redirect_uri', this.getGitlabRedirectUrl(req));
		params.append('response_type', 'code');
		params.append('scope', 'read_user');
		params.append('client_id', GITLAB_ID);
		return `${GITLAB_URL}/oauth/authorize?${params.toString()}`;
	}

	async getGitlabUserFromCode(req: Request, code: string): Promise<User> {
		const params = new URLSearchParams();
		params.append('redirect_uri', this.getGitlabRedirectUrl(req));
		params.append('client_id', GITLAB_ID);
		params.append('client_secret', GITLAB_SECRET);
		params.append('code', code);
		params.append('grant_type', 'authorization_code');
		const resp = await axios.post(
			`${GITLAB_API_URL}/oauth/token?${params.toString()}`,
			{}
		);
		const data: GitlabTokenResponse = resp.data;
		const userInfoResp = await axios.get(`${GITLAB_API_URL}/api/v4/user`, {
			headers: { Authorization: `Bearer ${data.access_token}` }
		});
		const userInfo: GitlabUserResponse = userInfoResp.data;
		let user = await this.findUser(userInfo.username, true);
		if (!user) user = await this.singupGitlab(userInfo, data);
		else user = await this.setGitlabTokens(user, data);
		return user;
	}

	async getGitlabUserInfo(
		req: Request,
		user: User
	): Promise<GitlabUserResponse | null> {
		try {
			const userInfoResp = await axios.get(
				`${GITLAB_API_URL}/api/v4/user`,
				{
					headers: { Authorization: `Bearer ${user.gitlabAT}` }
				}
			);
			return userInfoResp.data;
		} catch (e) {
			return null;
		}
	}

	async tryRefreshGitlabTokens(
		req: Request,
		user: User
	): Promise<User | null> {
		const params = new URLSearchParams();
		params.append('redirect_uri', this.getGitlabRedirectUrl(req));
		params.append('client_id', GITLAB_ID);
		params.append('client_secret', GITLAB_SECRET);
		params.append('refresh_token', user.gitlabRT);
		params.append('grant_type', 'refresh_token');
		try {
			const resp = await axios.post(
				`${GITLAB_API_URL}/oauth/token?${params.toString()}`,
				{}
			);
			const data: GitlabTokenResponse = resp.data;
			return this.setGitlabTokens(user, data);
		} catch (e) {
			return null;
		}
	}

	async verifyGitlabUser(req: Request, user: User): Promise<boolean> {
		let info = await this.getGitlabUserInfo(req, user);
		if (!info) {
			user = await this.tryRefreshGitlabTokens(req, user);
			if (!user) return false;
			info = await this.getGitlabUserInfo(req, user);
		}
		return info && info.username == user.name;
	}

	async deleteUser(user: User): Promise<void> {
		await this.revokeAll(user);
		await this.fsService.delete(await user.root, true);
		await this.userRepo.remove(user);
	}

	async changePassword(
		user: User,
		oldPW: string,
		newPw: string
	): Promise<void> {
		if (!(await argon2.verify(user.password, oldPW)))
			throw new ForbiddenException('Old password is wrong');
		user.password = await argon2.hash(newPw);
		await this.revokeAll(await this.userRepo.save(user));
	}
}
